-- TBD: Have configuration follow the action name?
ABT = LibStub("AceAddon-3.0"):NewAddon("ButtonTimers", "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("ButtonTimers", "enUS", true)
local LBF = LibStub("LibButtonFacade", true)

ABT_MaxButton = 12
ABT_MaxBar = 4
ABT_Widgets = "ABTWid"
local ABT_Bars = {}
local ABT_Buttons = {}
local currentBar = 1
local currentButton = 0
local outer_margin = 8
local ABT_TARGET = 1
local ABT_FOCUS = 2
local ABT_PLAYER = 3
local ABT_MOUSEOVER = 4
local ABT_PARTY = 5
local ABT_TARGET_NAMES = { [ABT_TARGET]="target", [ABT_FOCUS]="focus", [ABT_PLAYER]="player", [ABT_PARTY]="party" }
local ABT_AURA = 1
local ABT_COOLDOWN = 2
local ABT_NONE = 3
local ABT_VERTICAL = 1
local ABT_HORIZONTAL = 2
local ABT_TOPRIGHT = 1
local ABT_BOTTOMLEFT = 2
local ABT_ONBUTTONS = 3
local ABT_FULL = 1
local ABT_FIXED = 2
local TimeSinceLastUpdate = nil
local ABT_UpdateInterval = .05


function NS (arg1)
	if arg1 == nil then
		return "nil"
	elseif type(arg1) == "table" then
		local myval = "{"
		for i, j in pairs(arg1) do
			myval = myval.." ["..NS(i).."]="..NS(j)..","
		end
		myval = myval.."}"
		return myval
	else
		return tostring(arg1)
	end
end

function ABT:DebugTrace (arg1, enter)
	if ABT.db and ABT.db.char.showTrace == true then
		if (enter) then
			DEFAULT_CHAT_FRAME:AddMessage("ABT:"..arg1.." enter")
		else
			DEFAULT_CHAT_FRAME:AddMessage("ABT:"..arg1.." exit")
		end
	end
end

function ABT:DebugPrint (arg1)
	if ABT.db.char.showDebug == true then
		DEFAULT_CHAT_FRAME:AddMessage("ABT:"..arg1)
	end
end

function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end  -- if
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end  -- for
        return setmetatable(new_table, getmetatable(object))
    end  -- function _copy
    return _copy(object)
end  -- function deepcopy

-- parse a comma separated list into a table
function parseList (list)
	local _table = {}
	local idx = 1
	if list then
		for value in string.gmatch(list, "[^\n\r,]+") do
			tinsert(_table, value)
		end
	end
	return _table
end
function getKey(info)
	local key
	if #info == 3 then
		key = info[#info-2]..info[#info-1]..info[#info]
	elseif #info == 2 then
		key = info[#info-1]..info[#info]
	else
		key = info[#info]
	end
	return key
end

function ABT:GetValue (barIdx, buttonIdx, field)
	local key = nil
	local secondTalent = nil
	if buttonIdx and barIdx and field then
		key = "bar"..barIdx.."button"..buttonIdx..field
		if GetActiveTalentGroup() == 2 then -- if we're using an alt spec
			secondTalent = "bar"..barIdx.."button"..buttonIdx.."Alt" -- key that indicates settings for second spec
		end
	elseif barIdx and field then
		key = "bar"..barIdx..field
	elseif field then
		key = field
	end
	if key then
		if secondTalent and ABT.db.char[secondTalent] then
			if field == "timerColor" then
				ABT:DebugPrint ("key="..key.."secondary spec value "..NS(ABT.db.char[key.."2"]))
			end
			return ABT.db.char[key.."2"] -- alt spec setting exists
		end
		if field == "timerColor" then
			ABT:DebugPrint ("key="..key.."primary spec value "..NS(ABT.db.char[key]))
		end
		return ABT.db.char[key] -- primary spec setting
	end
end

function ABT:SetValue (barIdx, buttonIdx, field, value)
	local key = nil
	local secondTalent = nil
	ABT:DebugPrint ("SetValue ("..NS(barIdx)..", "..NS(buttonIdx)..", "..NS(field)..", "..NS(value)..")")

	if buttonIdx and barIdx and field then
		if GetActiveTalentGroup() == 2 then -- if we're using an alt spec
			secondTalent = "bar"..barIdx.."button"..buttonIdx.."Alt" -- key that indicates settings for second spec
			ABT:DebugPrint ("secondary spec activated "..secondTalent)
		end
		key = "bar"..barIdx.."button"..buttonIdx..field
	elseif barIdx and field then
		key = "bar"..barIdx..field
	elseif field then
		key = field
	end
	if key then
		if secondTalent then
			ABT:DebugPrint ("setting secondary spec value "..NS(key.."2").." = "..NS(value))
			ABT.db.char[key.."2"] = value
			ABT.db.char[secondTalent] = true
		else
			ABT.db.char[key] = value
		end
	end
end

function ABT:GetButtonLabel(info)
	local barIdx, buttonIdx = getIndex (info)
	return ABT:GetButtonName (barIdx, buttonIdx)
end

function ABT:GetButtonHidden(info)
	local barIdx, buttonIdx = getIndex (info)
	if barIdx and buttonIdx then
		return buttonIdx > ABT:GetValue (barIdx, nil, "buttonCount")
	else
		return false
	end
end

function getIndex(info)
	local buttonIdx = nil
	local barIdx = nil
	local field = nil

	for i = 1, #info do
		local a = string.match(info[i], "button(.+)")
		if a and tonumber(a) then
			buttonIdx = tonumber(a)
		else
			a = string.match(info[i], "bar(.+)")
			if a and tonumber(a) then
				barIdx = tonumber(a)
			else
				field = info[i]
			end
		end
	end
	return barIdx, buttonIdx, field
end

function ABT:getFunc(info)
	local barIdx, buttonIdx, field = getIndex(info)
	return ABT:GetValue (barIdx, buttonIdx, field)
end

function ABT:setFunc(info, value)
	local barIdx, buttonIdx, field = getIndex(info)
	ABT:SetValue (barIdx, buttonIdx, field, value)

	local barIdx, buttonIdx = getIndex(info)
	if barIdx ~= nil then   -- button option or bar option
		ABT:SetBar (barIdx)
	else                                             -- general option
		for barIdx = 1, ABT_MaxBar do
			ABT:SetBar (barIdx)
		end
	end
	ABT:GetDebuffs()
end

local buttonOptions = {
	type = "group",
	name = function(info) return ABT:GetButtonLabel(info) end,
	hidden = function(info) return ABT:GetButtonHidden(info) end,
	desc = "Options for each button",
	childGroups = "tree",
	inline = false,
	args = {
		target = {
			order = 2,
			type = "select",
			name = L["Target"],
			desc = L["The target on which this button will operate."],
			values = { [ABT_TARGET]=L["Target"], [ABT_FOCUS]=L["Focus"], [ABT_PLAYER]=L["Player"], [ABT_PARTY]=L["Party/Raid Member"] },
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		playerTarget = {
			order = 3,
			type = "select",
			hidden = function (info)  local barIdx, buttonIdx = getIndex(info) return ABT:GetValue (barIdx, buttonIdx, "target") ~= ABT_PARTY  end,
			name = L["Party/Raid Member"],
			desc = L["The party / raid member to target"],
			values = function () return ABT:GetPartyList() end,
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		timerType = {
			type = "select",
			order = 10,
			name = L["Type"],
			values = { [ABT_AURA]=L["Aura"], [ABT_COOLDOWN]=L["Cooldown"], [ABT_NONE]=L["None"]},
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		barType = {
			type = "select",
			order = 20,
			name = L["Timer Type"],
			desc = L["Choose full time to see the full length of the timer, choose fixed time to see a set maximum time."],
			values = { [ABT_FULL]=L["Full time"], [ABT_FIXED]=L["Fixed time"] },
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		barTime = {
			order = 30,
			type = "range",
			hidden = function (info) local barIdx, buttonIdx = getIndex(info) return ABT:GetValue (barIdx, buttonIdx, "barType") ~= ABT_FIXED  end,
			name = L["Bar Time"],
			desc = L["Max time displayed on bar"],
			min = 1,
			max = 100,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},
		showOthers = {
			order = 40,
			type = "toggle",
			hidden = function (info)  local barIdx, buttonIdx = getIndex(info) return (ABT:GetValue (barIdx, buttonIdx, "timerType") ~= ABT_AURA)  end,
			name = L["Show others spells"],
			desc = L["Show this spell on the target even if someone else cast it."],
			set = "setFunc",
			get = "getFunc",
		},
		auras = {
			order = 50,
			hidden = function (info)  local barIdx, buttonIdx = getIndex(info) return ABT:GetValue (barIdx, buttonIdx, "timerType") ~= ABT_AURA  end,
			type = "input",
			name = L["Other Auras"],
			desc = L["Add any additional auras you want the timer to show as a comma separated list."],
			multiline = true,
			set = "setFunc",
			get = "getFunc",
		},
		spell = {
			order = 60,
			hidden = function (info)  local barIdx, buttonIdx = getIndex(info) return ABT:GetValue (barIdx, buttonIdx, "timerType") ~= ABT_COOLDOWN  end,
			type = "input",
			name = L["Cooldown Spell"],
			desc = L["If you would like to see the cooldown for a spell other than the one on the button, enter it here."],
			set = "setFunc",
			get = "getFunc",
		},
		timerColor = {
			order = 70,
            name = L["Bar Color"],
            desc = L["Bar color for timers"],
            type = "color",
			hasAlpha = true,
            set = function (info, r, g, b, a) return ABT:setFunc (info, {r, g, b, a}) end,
            get = function (info) return unpack(ABT:getFunc(info)) end
        },
		timerWarnColor = {
			order = 70,
 			hidden = function (info) local barIdx, buttonIdx = getIndex(info) 
									return ABT:GetValue (barIdx, buttonIdx, "timerType") ~= ABT_AURA or 
									ABT:GetValue (barIdx, buttonIdx, "colorChange") ~= true end,
           name = L["Warning Bar Color"],
            desc = L["Bar color for timers when remaining time < cast time"],
            type = "color",
			hasAlpha = true,
            set = function (info, r, g, b, a) return ABT:setFunc (info, {r, g, b, a}) end,
            get = function (info) return unpack(ABT:getFunc(info)) end
        },
		colorChange = {
			order = 65,
			hidden = function (info)  	local barIdx, buttonIdx = getIndex(info) 
										return ABT:GetValue (barIdx, buttonIdx, "timerType") ~= ABT_AURA  end,
			type = "toggle",
			name = L["Warn < cast time."],
			desc = L["Change the bar color when timer < cast time."],
			set = "setFunc",
			get = "getFunc",
		},
		timerAdjust = {
			order = 90,
			type = "input",
			name = L["Adjust Timer"],
			desc = L["Number of seconds to add to timer (can be negative)."],
			set = "setFunc",
			get = "getFunc",
		}
	}
}

function ABT:GetButtonOptions (button)
	local _options = deepcopy (buttonOptions)
	_options.order = button
	return _options
end

function ABT:GetBarDesc(info)
	local offset = ABT:getFunc(info)
	local barIdx = math.modf(offset / 12) + 1
	local buttonIdx = (offset % 12) + 1
--  ABT:DebugPrint( "First button of bar is: Action bar: "..NS(barIdx)..", button: "..NS(buttonIdx))
	return "Bar: "..NS(barIdx)..", Button: "..NS(buttonIdx)
end

local barOptions = {
	type = "group",
	name = "Bar",
--	handler = ABT,
	desc = "Options for the bar",
	childGroups = "tree",
	inline = false,
	args = {
		enabled = {
			order = 10,
			type = "toggle",
			name = L["Bar Enabled"],
			desc = L["Enable this bar"],
			set = "setFunc",
			get = "getFunc",
		},
		inCombat = {
			order = 15,
			type = "toggle",
			name = L["Hide out of combat"],
			desc = L["Show this bar only when in combat"],
			set = "setFunc",
			get = "getFunc",
		},
		hideTooltips={
			order = 18,
			name= L["Hide Tooltips"],
			desc= L["Hide tooltips on buttons"],
			type= "toggle",
			set = "setFunc",
			get = "getFunc",
		},
		locked = {
			order = 19,
			type = "toggle",
			name = L["Bar Locked"],
			desc = L["Lock bar in place"],
			set = "setFunc",
			get = "getFunc",
		},
		buttonCount = {
			order = 20,
			type = "range",
			name = L["Button Count"],
			desc = L["Number of Buttons on bar"],
			min = 1,
			max = 12,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},
		actionDesc = {
			order = 29,
			type = "description",
			name = L["First button on bar is action slot:"],
		},
		actionOffset = {
			order = 30,
			type = "range",
			name = function(info) return ABT:GetBarDesc(info) end,
			desc = L["Move this slider to change which action slots are shown on the bar."],
			min = 0,
			max = 119,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},	
		orientation = {
			order = 40,
			type = "select",
			name = L["Orientation"],
			values = { [ABT_VERTICAL]=L["Vertical"], [ABT_HORIZONTAL]=L["Horizontal"] },
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		location = {
			order = 45,
			type = "select",
			name = L["Timer Location"],
			values = { [1]=L["Right/Top"], [2]=L["Left/Bottom"], [3]=L["Timers on Buttons"] },
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		spacing = {
			order = 60,
			type = "range",
			name = L["Button Spacing"],
			desc = L["The space between the buttons"],
			min = 0,
			min = 0,
			max = 16,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},
		scale = {
			order = 70,
			type = "range",
			name = L["Bar Scale"],
			desc = L["Allows you to scale the bar size"],
			min = .1,
			max = 2,
			step = .01,
			bigStep = .01,
			set = "setFunc",
			get = "getFunc",
		},
		length = {
			order = 80,
			type = "range",
			name = L["Bar Length"],
			desc = L["Length of bar in pixels"],
			min = 1,
			max = 600,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},
		button1 = ABT:GetButtonOptions(1),
		button2 = ABT:GetButtonOptions(2),
		button3 = ABT:GetButtonOptions(3),
		button4 = ABT:GetButtonOptions(4),
		button5 = ABT:GetButtonOptions(5),
		button6 = ABT:GetButtonOptions(6),
		button7 = ABT:GetButtonOptions(7),
		button8 = ABT:GetButtonOptions(8),
		button9 = ABT:GetButtonOptions(9),
		button10 = ABT:GetButtonOptions(10),
		button11 = ABT:GetButtonOptions(11),
		button12 = ABT:GetButtonOptions(12),
	}
}

function ABT:GetBarOptions (bar)
	local _options = deepcopy (barOptions)
	_options.name = "Bar ".. bar
	_options.order = bar
	return _options
end

local options = {
    name = "ButtonTimers",
    handler = ABT,
    type = 'group',
	childGroups = "tree",
    args = {
		enable={
			order = 10,
			name= L["Enable"],
			desc= L["Enables / disables the addon"],
			type= "toggle",
			set = "setFunc",
			get = "getFunc",
		},
	    showDebug = {
			order = 100,
			hidden = true,
            name = L["ShowDebug"],
            desc = L["Show debug messages"],
            type = "toggle",
            set = "setFunc",
            get = "getFunc",
        },
	    showTrace = {
			order = 110,
			hidden = true,
            name = L["ShowTrace"],
            desc = L["Show trace messages"],
            type = "toggle",
            set = "setFunc",
            get = "getFunc",
        },
		showBackground = {
			order = 120,
			hidden = true,
            name = L["ShowBackground"],
            desc = L["Show background bar"],
            type = "toggle",
            set = "setFunc",
            get = "getFunc",
        },
		bar1 = ABT:GetBarOptions(1),
		bar2 = ABT:GetBarOptions(2),
		bar3 = ABT:GetBarOptions(3),
		bar4 = ABT:GetBarOptions(4),
    },
}

function ABT:InitDefaults()
	local defaults = {
		char = {}
	}
	defaults.char.enable = true
	defaults.char.showDebug = false
	defaults.char.showTrace = false
	defaults.char.showBackground = false
	for bar = 1, ABT_MaxBar do
		defaults.char["bar"..bar.."enabled"] = true
		defaults.char["bar"..bar.."locked"] = false
		defaults.char["bar"..bar.."orientation"] = ABT_VERTICAL
		defaults.char["bar"..bar.."location"] = ABT_TOPRIGHT
		defaults.char["bar"..bar.."scale"] = 1
		defaults.char["bar"..bar.."length"] = 300
		defaults.char["bar"..bar.."spacing"] = 8
		defaults.char["bar"..bar.."buttonCount"] = 12
		defaults.char["bar"..bar.."actionOffset"] = (10-bar) * 12 -- default bars 10, 9, 8, 7
		defaults.char["bar"..bar.."auraColor"] = { 0, 1, 1, .5 }
		defaults.char["bar"..bar.."cooldownColor"] = { 0, 1, 0, .5 }
		defaults.char["bar"..bar.."hideTooltips"] = false
		for button = 1, ABT_MaxButton do
			defaults.char["bar"..bar.."button"..button.."target"] = ABT_TARGET
			defaults.char["bar"..bar.."button"..button.."timerType"] = ABT_AURA
			defaults.char["bar"..bar.."button"..button.."barType"] = ABT_FULL
			defaults.char["bar"..bar.."button"..button.."barTime"] = 20
			defaults.char["bar"..bar.."button"..button.."showOthers"] = false
			defaults.char["bar"..bar.."button"..button.."auras"] = nil
			defaults.char["bar"..bar.."button"..button.."spell"] = nil
			defaults.char["bar"..bar.."button"..button.."colorChange"] = nil
			defaults.char["bar"..bar.."button"..button.."timerColor"] = { 0, 1, 1, .5 }
			defaults.char["bar"..bar.."button"..button.."timerWarnColor"] = { 1, 0, 0, .5 }
			defaults.char["bar"..bar.."button"..button.."timerAdjust"] = "0"
			-- alt spec inits
			defaults.char["bar"..bar.."button"..button.."Alt"] = nil
			defaults.char["bar"..bar.."button"..button.."target2"] = ABT_TARGET
			defaults.char["bar"..bar.."button"..button.."timerType2"] = ABT_AURA
			defaults.char["bar"..bar.."button"..button.."barType2"] = ABT_FULL
			defaults.char["bar"..bar.."button"..button.."barTime2"] = 20
			defaults.char["bar"..bar.."button"..button.."showOthers2"] = false
			defaults.char["bar"..bar.."button"..button.."auras2"] = nil
			defaults.char["bar"..bar.."button"..button.."spell2"] = nil
			defaults.char["bar"..bar.."button"..button.."colorChange2"] = nil
			defaults.char["bar"..bar.."button"..button.."timerColor2"] = { 0, 1, 1, .5 }
			defaults.char["bar"..bar.."button"..button.."timerWarnColor2"] = { 1, 0, 0, .5 }
			defaults.char["bar"..bar.."button"..button.."timerAdjust2"] = "0"
		end
	end
	return defaults
end

function ABT:OnInitialize()
  -- Code that you want to run when the addon is first loaded goes here.
	ABT:DebugTrace ("OnInitialize", true)
	ABT:RegisterChatCommand("buttontimers", "SlashProcessorFunc")
	-- bt overwrites bartender, doh
	ABT:RegisterChatCommand("abt", "SlashProcessorFunc")
    LibStub("AceConfig-3.0"):RegisterOptionsTable("ButtonTimers", options)
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ButtonTimers", "ButtonTimers")
    self.db = LibStub("AceDB-3.0"):New("ButtonTimersDB", ABT:InitDefaults(), "Default")
	ABT:MakeWidgets()
	ABT:DebugTrace ("OnInitialize")

	-- Binding Variables
	for bar=1,ABT_MaxBar do
		setglobal("BINDING_HEADER_ButtonTimersBar"..bar, "ButtonTimers Bar "..bar);
		for button=1, ABT_MaxButton do
			local buttonName = ABT_Widgets.."bar"..bar.."button"..button
			setglobal("BINDING_NAME_CLICK "..buttonName..":LeftButton", "Button "..button);
		end
	end
end

-- Button Facade Callback
function ABT:OnSkin(skin, gloss, backdrop, group, button, colors)
	if group then
		ABT:DebugPrint ("OnSkin group "..NS(group).." skin="..NS(skin))
		ABT.db.char["buttonFacade"..group] = {skin, gloss, backdrop, colors}
	end
end

function ABT:OnEnable()
	ABT:DebugTrace ("OnEnable", true)
    -- Called when the addon is enabled
	self:RegisterEvent ("UNIT_AURA", "GetDebuffs")
	self:RegisterEvent ("PLAYER_TARGET_CHANGED", "GetDebuffs")
	self:RegisterEvent ("PLAYER_FOCUS_CHANGED", "GetDebuffs")
	self:RegisterEvent ("PLAYER_ENTERING_WORLD", "SetAllBars")
	self:RegisterEvent ("PLAYER_TALENT_UPDATE", "SetAllBars")
	self:RegisterEvent ("ACTIONBAR_UPDATE_COOLDOWN", "GetCooldowns")
	self:RegisterEvent ("ACTIONBAR_UPDATE_USABLE", "GetCooldowns")
	self:RegisterEvent ("PARTY_MEMBERS_CHANGED", "UpdatePlayerTargets")
	self:RegisterEvent ("ACTIONBAR_SLOT_CHANGED", function (event, arg1) 
													ActionButton_OnEvent(event) ABT:ButtonUpdate() 
													end)
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "OnEnterCombat")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "OnLeaveCombat")
	ABT:RegisterBF()
	ABT:ButtonUpdate()
	ABT:DebugTrace ("OnEnable")
end

function ABT:OnEnterCombat ()
	for barIdx = 1, ABT_MaxBar do
		local barEnabled = ABT:GetValue (barIdx, nil, "enabled")
		local inCombat = ABT:GetValue (barIdx, nil, "inCombat")
		if barEnabled and inCombat then
			ABT_Bars[barIdx]:Show()
		end
	end
end

function ABT:OnLeaveCombat ()
	for barIdx = 1, ABT_MaxBar do
		local barEnabled = ABT:GetValue (barIdx, nil, "enabled")
		local inCombat = ABT:GetValue (barIdx, nil, "inCombat")
		if barEnabled and inCombat then
			ABT_Bars[barIdx]:Hide()
		end
		if ABT_Bars[barIdx].needsRefresh then
			ABT:SetBar(barIdx)
		end
	end
end

function ABT:ButtonUpdate ()
	ABT:GetCooldowns()
	ABT:GetDebuffs()
end

function ABT:OnDisable()
	ABT:DebugTrace ("OnDisable", true)
    -- Called when the addon is disabled
	ABT:DebugTrace ("OnDisable")
end

function ABT:SlashProcessorFunc(input)
	ABT:DebugTrace ("SlashProcessorFunc", true)
  -- Process the slash command ('input' contains whatever follows the slash command)
	ABT:DebugPrint ("Received command "..input)
	if (input == "reset") then
		for barIdx = 0, ABT_MaxButton do
			local bar = ABT_Bars[barIdx]
			if bar then
				bar:ClearAllPoints()
				bar:SetPoint("CENTER")
			end
		end
	elseif (input == "toggleDebug") then
		if ABT.db.char.showDebug then
			ABT:DebugPrint ("turning debug messages off")
			ABT.db.char.showDebug = false
		else
			ABT.db.char.showDebug = true
			ABT:DebugPrint ("turning debug messages on")
		end
	elseif (input == "toggleTrace") then
		if ABT.db.char.showTrace then
			ABT:DebugPrint ("turning trace messages off")
			ABT.db.char.showTrace = false
		else
			ABT.db.char.showTrace = true
			ABT:DebugPrint ("turning trace messages on")
		end
	else
		InterfaceOptionsFrame_OpenToCategory("ButtonTimers")
	end
	ABT:DebugTrace ("SlashProcessorFunc")
end

-- update the bars
function ABT:OnUpdate (self, elapsed)
	if self:IsShown() then --and ABT:GetValue(nil, nil, "enable") then
		if self.TimeSinceLastUpdate == nil then 
			self.TimeSinceLastUpdate = elapsed
		end
		self.TimeSinceLastUpdate = self.TimeSinceLastUpdate + elapsed; 	
		if (self.TimeSinceLastUpdate > ABT_UpdateInterval) then
			local currentTime = GetTime()
			self.TimeSinceLastUpdate = 0;

			if self.expirationTime and self.expirationTime > currentTime then
				local remaining = self.expirationTime - currentTime;
				if remaining > 10 then
					if self.timer.useMe == true then
						self.timer:SetFormattedText(SecondsToTimeAbbrev(remaining))
					else
						self.countdown:SetFormattedText(SecondsToTimeAbbrev(remaining))
					end
				else
					if self.timer.useMe == true then
						self.timer:SetFormattedText("%.1f", remaining)
					else
						self.countdown:SetFormattedText("%.1f", remaining)
					end
				end
				if self.castTime and self.tex.warnColor and remaining * 1000 < self.castTime then
--				ABT:DebugPrint (NS(self.castTime).." "..NS(self.tex.warnColor).." "..NS(remaining))
					self.tex:SetVertexColor (unpack(self.tex.warnColor))
				end
				ABT:SetStatusBarRemaining (self, remaining)
			else
				self.expirationTime = nil
				self.countdown:SetText(nil)
				self.timer:SetText(nil)
				self.spellName:SetText(nil)
				self.status:Hide()
			end
		end
	end
end

function ABT:GetButtonName (barIdx, buttonIdx)
	local buttonName
	local button = ABT_Buttons[barIdx][buttonIdx]
	local actionSlot = button:GetAttribute("action")	
	local actiontype, id, subType
	if actionSlot then
		actiontype, id, subType = GetActionInfo(actionSlot)
--		ABT:DebugPrint ("type="..NS(actiontype).." id="..NS(id).."subType="..NS(subType))
	end
	if actiontype == "spell" and id > 0 then
		local spellName = GetSpellName (id, BOOKTYPE_SPELL)
--		ABT:DebugPrint ("spellName = "..NS(spellName))
		if spellName then
			buttonName = spellName
		end
	elseif actiontype == "macro" then
		buttonName = GetActionText(actionSlot)
	elseif actiontype == "item" then
		buttonName = GetItemInfo(id)
	end
	if buttonName == nil then
		buttonName = "bar"..barIdx.."button"..buttonIdx
	end
	return buttonName
end

--
-- Help, does this really need to do all buttons for each call?
-- couldn't we just update cooldowns on buttons that gave the action?
function ABT:GetCooldowns (self, arg1)
	ABT:DebugTrace ("GetCooldowns", true)
	ABT:InitButtonLoop()
	repeat 
		local barIdx, buttonIdx = ABT:GetNextButton()
		if barIdx and buttonIdx then
			local key = "bar"..barIdx.."button"..buttonIdx
			local timerType = ABT:GetValue (barIdx, buttonIdx, "timerType")
			if timerType == ABT_COOLDOWN then -- cooldown timer
				local otherSpell = ABT:GetValue (barIdx, buttonIdx, "spell")
				local button = ABT_Buttons[barIdx][buttonIdx]
				local slot = button:GetAttribute ("action")

				local start, duration, enabled
				if otherSpell then
					start, duration, enabled = GetSpellCooldown (otherSpell)
				elseif (slot) then
					start, duration, enabled = GetActionCooldown (slot)
--					DEFAULT_CHAT_FRAME:AddMessage("slot "..slot.." duration="..duration)

				end
				if start and duration then
					-- we have a cooldown to track
					if start > 0 and duration > 1.5 then
						local add = tonumber(ABT:GetValue (barIdx, buttonIdx, "timerAdjust"))
						button.expirationTime = start + duration + (add or 0)
						if (button.expirationTime <= 0) then
							button.expirationTime = nil
						else
							if ABT:GetValue (barIdx, nil, "orientation") == ABT_HORIZONTAL then
								button.spellName:SetText(nil)
							elseif otherSpell then
								button.spellName:SetText(otherSpell)
							else
								-- find the name of the spell, so we can put it on the timer
								spellName = ABT:GetButtonName (barIdx, buttonIdx)
								if spellName then
									button.spellName:SetText(spellName)
								end
							end
							local barTime
							if ABT:GetValue (barIdx, buttonIdx, "barType") == ABT_FIXED then
								barTime = ABT:GetValue (barIdx, buttonIdx, "barTime")
							else
								barTime = duration
							end
							button.tex.max = barTime
							if button.timer.useMe == false then
								button.status:Show()
							end
						end
					-- we have a cooldown timer running, but the cooldown is no longer
					-- active. Reset it. If it's less than the GCD, though, just let it run out.
					elseif button.expirationTime and (button.expirationTime - GetTime()) > 1.5 then
						button.expirationTime = nil
					end
				end
			end
		end
	until buttonIdx == nil
	ABT:DebugTrace ("GetCooldowns")
end

function ABT:GetDebuffs(event, arg1)
	ABT:DebugTrace ("GetDebuffs", true)
	local unit = nil
--	ABT:DebugPrint (NS(event).." "..NS(arg1))
	if (event == "UNIT_AURA" and arg1 == "target") or event == "PLAYER_TARGET_CHANGED" then
		unit = "target"
	elseif (event == "UNIT_AURA" and arg1 == "focus") or event == "PLAYER_FOCUS_CHANGED" then
		unit = "focus"
	elseif (event == "UNIT_AURA") then
		unit = arg1
	end
	-- loop over buttons looking for aura timer buttons matching the target
	local barIdx = nil
	local buttonIdx = nil
	ABT:InitButtonLoop()
	repeat 
		barIdx, buttonIdx = ABT:GetNextButton()
--			ABT:DebugPrint ("bar="..NS(barIdx).." button="..NS(buttonIdx))
		if barIdx and buttonIdx then
			local button = ABT_Buttons[barIdx][buttonIdx]
			local key = "bar"..barIdx.."button"..buttonIdx
			local timerType = ABT:GetValue (barIdx, buttonIdx, "timerType")
			local target = button:GetAttribute ("unit")
			if (target == nil) then target = "target" end
			local showOthers = ABT:GetValue (barIdx, buttonIdx, "showOthers")
			local buttonFound = false -- haven't yet found a match for this button
--				ABT:DebugPrint ("timerType="..NS(timerType).." target="..NS(target).." unit="..NS(unit))
			if timerType == ABT_AURA and (target == unit or unit == nil) then -- aura timer
				local effectList = parseList (ABT:GetValue (barIdx, buttonIdx, "auras"))
--					ABT:DebugPrint ("button "..buttonIdx.." on bar "..barIdx.." looking for target "..NS(target).." buttonSpell="..NS(button.spell))
				-- build effects list
				if button.action then
					local actionType, actionId, actionSubType = GetActionInfo (button.action)
					if actionType == "spell" and actionId > 0 then
						local spellName, _ = GetSpellName ( actionId, BOOKTYPE_SPELL )
						effectList[100] = spellName
					end
				end
				-- now look for the aura for this button
				local effect
				for _, effect in pairs(effectList) do			
--						ABT:DebugPrint ("Looking for "..effect.." on button "..button:GetName().." target="..NS(target))
					local name, rank, iconTexture, count, debuffType, duration, timeLeft, expirationTime, isMine, caster
					timeLeft = nil
					isMine = nil
					expirationTime = nil
					name, rank, iconTexture, count, debuffType, duration, expirationTime, caster = UnitDebuff (target, effect); -- 3.0

					if name == nil then -- wasn't found as a debuff, try it as a buff
						name, rank, iconTexture, count, debuffType, duration, expirationTime, caster = UnitBuff (target, effect); -- 3.0
					end
					-- the effect was found on the mob, now lets see if it's ours
					-- use the effect if it is ours, or if the anyCaster flag is set and the button 
					-- was not previously found. This will prioritize our own spells for display if multiple
					-- matching effects are found
					isMine = (caster == "player" or caster == "pet") -- 3.1 
					if name ~= nil and (isMine or (showOthers and not buttonFound)) then
--							ABT:DebugPrint ("found effect "..effect)
						local currenttime = GetTime()
						timeLeft = expirationTime-currenttime
						local timeStart = expirationTime - duration
						local add = tonumber(ABT:GetValue (barIdx, buttonIdx, "timerAdjust"))
						button.expirationTime = expirationTime + (add or 0)-- time left on this effect
						if (button.expirationTime <= 0) then
							button.expirationTime = nil
						else
							button.count = count
	--							ABT:DebugPrint ("expirationTime ="..button.expirationTime)
							local barTime
							if ABT:GetValue (barIdx, buttonIdx, "barType") == ABT_FIXED then
								barTime = ABT:GetValue (barIdx, buttonIdx, "barTime")
							else
								barTime = duration
							end
	--							button.status:SetMinMaxValues (0, barTime)
							button.tex.max = barTime
							button.tex:SetVertexColor (unpack(button.tex.timerColor))
							if button.tex.warnColor then
								button.castTime = ABT:GetCastTime (button)
							end
							if button.timer.useMe == false then
								button.status:Show()
							end
							if ABT:GetValue (barIdx, nil, "orientation") == ABT_HORIZONTAL then
								if count > 1 then
									button.spellName:SetText("x"..NS(count))
								else
									button.spellName:SetText(nil)
								end
							elseif target ~= "target" then
								button.spellName:SetText(effect..": "..NS(UnitName(target)))
							elseif count > 1 then
								button.spellName:SetText(effect.." x"..NS(count))
							else
								button.spellName:SetText(effect)
							end
							buttonFound = true
						end
					end
				end
				if not buttonFound then
					button.expirationTime = nil
				end
			end
		end
	until buttonIdx == nil
	ABT:DebugTrace ("GetDebuffs")
end

function ABT:InitButtonLoop()
	currentBar = 1
	currentButton = 0
end

function ABT:GetNextButton()
	while true do
		currentButton = currentButton + 1
		if currentButton > ABT_MaxButton then
			currentBar = currentBar + 1
			currentButton = 1
			if currentBar > ABT_MaxBar then
				return nil, nil
			end
		end
		local barEnabled = self.db.char["bar"..currentBar.."enabled"]
		local barButtonCount = self.db.char["bar"..currentBar.."buttonCount"]
		if barEnabled and currentButton <= barButtonCount then
			return currentBar, currentButton
		end
	end
end

function ABT:StartMoving (self)
	ABT:DebugTrace ("StartMoving", true)
	if (not InCombatLockdown() and not self.locked) then
		self:StartMoving()
	end
	ABT:DebugTrace ("StartMoving")
end

function ABT:StopMoving (self)
	ABT:DebugTrace ("StopMoving", true)
	self:StopMovingOrSizing ()
	local barIdx = self.barIdx
	local attach = { self:GetPoint() }
	ABT:SetValue (barIdx, nil, "attach", attach)
	ABT:DebugPrint ("stop moving bar="..barIdx.." "..NS(attach))
	ABT:DebugTrace ("StopMoving", true)
end

function ABT:MakeWidgets()
	ABT:DebugTrace ("MakeWidgets", true)
	for barIdx = 1, ABT_MaxBar do
		local barName = ABT_Widgets.."bar"..barIdx
		local bar = CreateFrame ("Frame", barName, UIParent)
--		local bar = CreateFrame ("Frame", barName)
		ABT_Bars[barIdx] = bar
		bar.barIdx = barIdx
		-- this texture is handy for seeing where the frame is to drag it.
		-- only show it if the bar is draggable.
		bar.texture = bar:CreateTexture ()
		bar.texture:SetAllPoints(bar)
		bar.texture:SetTexture(1, 1, 1, .1)
		bar.texture:Hide()
		--
		-- setup for dragging
		bar:SetMovable (true)
		bar:SetClampedToScreen (true)
		bar:EnableMouse (true)
		bar:SetScript ("OnMouseDown", function (self) ABT:StartMoving (self) end)
		bar:SetScript ("OnMouseUp", function (self) ABT:StopMoving (self) end)
		bar:SetScript ("OnDragStop", function (self) ABT:StopMoving () end)

		ABT_Buttons[barIdx] = {}
		for buttonIdx = 1, ABT_MaxButton do
			local buttonName = barName.."button"..buttonIdx
			local button = CreateFrame ("CheckButton", buttonName , bar, "ABTButtonTemplate")
			ABT_Buttons[barIdx][buttonIdx] = button
			button:SetScript ("OnUpdate", function (self, elapsed) ActionButton_OnUpdate(self, elapsed) ABT:OnUpdate (self, elapsed) end)
			button.barIdx = barIdx
			button.buttonIdx = buttonIdx
			button.status = _G[buttonName.."StatusBar"]
			button.tex = _G[buttonName.."StatusBarTex"]
			button.spellName = _G[buttonName.."StatusBarSpellName"]
			button.countdown = _G[buttonName.."StatusBarCountdown"]
			button.timer = _G[buttonName.."Timer"]
			button.status:Show()
			button.cooldown = _G[buttonName.."Cooldown"]
			button:HookScript ("OnEnter", 
									function () 
										if ABT:GetValue(barIdx, nil, "hideTooltips") then 
											GameTooltip:Hide() 
										end 
									end)
		end
		ABT:SetBar(barIdx)
	end
	ABT:DebugTrace ("MakeWidgets")
end

-- register buttons with buttonfacade
function ABT:RegisterBF (event, arg1)
	ABT:DebugTrace ("RegisterBF", true)
	if LBF then
		-- buttonfacade callback
		LBF:RegisterSkinCallback("ButtonTimers", self.OnSkin, self)
		for barIdx = 1, ABT_MaxBar do
			local groupID = "Button Timers Bar"..barIdx
			local group = LBF:Group("ButtonTimers", groupID)
			local bfOptions = ABT.db.char["buttonFacade"..groupID]
			if bfOptions then
				group.SkinID,group.Gloss,group.Backdrop,group.Colors = unpack(bfOptions)
			end

			local bar = ABT_Bars[barIdx]

			for buttonIdx = 1, ABT_MaxButton do
				local buttonName = ABT_Widgets.."bar"..barIdx.."button"..buttonIdx
				local button = ABT_Buttons[barIdx][buttonIdx]
				-- ButtonFacade Support
				local btnData = {
					Icon = _G[buttonName.."Icon"],
					Cooldown = _G[buttonName.."Cooldown"],
					HotKey = _G[buttonName.."HotKey"],
					Count = _G[buttonName.."Count"],
					Name = _G[buttonName.."Name"],
				}
				group:AddButton(button, btnData)
			end
		end
	end
	ABT:DebugTrace ("RegisterBF")
end

function ABT:ButtonVisible (barIdx, buttonIdx)
	local buttonCount = ABT:GetValue (barIdx, nil, "buttonCount")
	return buttonIdx <= buttonCount
end

local scaleInit = {}

function ABT:SetAllBars (event, arg1)
	for barIdx = 1, ABT_MaxBar do
		ABT:SetBar (barIdx)
	end
end

function ABT:SetBar (barIdx)
	ABT:DebugTrace ("SetBar"..barIdx, true)
	
	if InCombatLockdown() then
		-- we're in combat, don't do anything now, but flag it so we
		-- can do the operation when we leave combat.
		ABT_Bars[barIdx].needsRefresh = true
		return
	end
	local width = 0
	local height = 0

	local buttonCount = ABT:GetValue(barIdx, nil, "buttonCount")
	local orientation = ABT:GetValue(barIdx, nil, "orientation")
	local location = ABT:GetValue(barIdx, nil, "location") 
	local scale = ABT:GetValue(barIdx, nil, "scale")
	local length = ABT:GetValue(barIdx, nil, "length")
	local margin = ABT:GetValue(barIdx, nil, "spacing")
	local actionOffset = ABT:GetValue(barIdx, nil, "actionOffset")
	local enabled = ABT:GetValue(barIdx, nil, "enabled") and ABT:GetValue(nil, nil, "enable")
	local locked = ABT:GetValue(barIdx, nil, "locked")
	local showBackground = ABT:GetValue(nil, nil, "showBackground")
	local inCombat = ABT:GetValue(barIdx, nil, "inCombat")
	
	local bar = ABT_Bars[barIdx]

	-- bar enable
	if enabled then
		bar:Show()
	else
		bar:Hide()
		return
	end
	
	-- bar lockdown
	bar.locked = locked
	
	local attach = ABT:GetValue(barIdx, nil, "attach")
--	ABT:DebugPrint ("read location from DB "..NS(attach))
	-- set bar position
	if (attach == nil) then
		attach = {"CENTER", nil, "CENTER", (44 * (barIdx-1)), 0}
		ABT:SetValue (barIdx, nil, "attach", attach)
--		ABT:DebugPrint ("initialized attach="..NS(attach))
	end
 	bar:ClearAllPoints()
	bar:SetPoint(unpack(attach))

	-- scale option
	local oldscale = bar:GetScale()
	local newscale = scale  * .6

	if oldscale ~= newscale and attach then
		bar:SetScale(newscale)
		if scaleInit[barIdx] ~= nil then
			local point, parent, relpoint, xOfs, yOfs = unpack(attach)
			xOfs = xOfs * oldscale / newscale
			yOfs = yOfs * oldscale / newscale
			attach = {point, parent, relpoint, xOfs, yOfs}
			ABT:SetValue (barIdx, nil, "attach", attach)
--			ABT:DebugPrint ("scaled DB attach="..NS(attach))
			bar:ClearAllPoints()
			bar:SetPoint(unpack(attach))
		else
			scaleInit[barIdx] = true		
		end
	end
	
	-- button count, target
	for buttonIdx = 1, ABT_MaxButton do
		local button = ABT_Buttons[barIdx][buttonIdx]
		if ABT:ButtonVisible (barIdx, buttonIdx) then
			button:Show()
			local targetIdx = ABT:GetValue(barIdx, buttonIdx, "target")
			local targetUnit = nil
			if targetIdx == ABT_PARTY then
				local playerTarget = ABT:GetValue(barIdx, buttonIdx, "playerTarget")
				targetUnit = ABT:FindPlayerUnitId (playerTarget)
			elseif targetIdx ~= ABT_TARGET then
				targetUnit = ABT_TARGET_NAMES[targetIdx]
			else
				targetUnit = nil
			end
			button:SetAttribute("unit", targetUnit)

			local statusbarTexture = "Interface\\PaperDollInfoFrame\\UI-Character-Skills-Bar"
			local statusbarTexture = "Interface\\TargetingFrame\\UI-StatusBar"
			local timerColor = ABT:GetValue (barIdx, buttonIdx, "timerColor")
			local warnColor = ABT:GetValue (barIdx, buttonIdx, "timerWarnColor")
			local colorChange = ABT:GetValue (barIdx, buttonIdx, "colorChange")
			local timerType = ABT:GetValue (barIdx, buttonIdx, "timerType")
			button.tex.timerColor = timerColor
--			ABT:DebugPrint ("colorChange="..NS(colorChange).." warnColor="..NS(warnColor).." timerType="..NS(timerType))
			if colorChange and warnColor and timerType == ABT_AURA then
--				ABT:DebugPrint("Setting warning color on: "..button:GetName())
				button.tex.warnColor = warnColor
			else
				button.tex.warnColor = nil
			end
			button.tex:SetVertexColor (unpack(timerColor))

--			ABT:DebugPrint ("setting unit for "..button:GetName().." to "..NS(targetUnit).." targetIdx="..targetIdx)
			button:SetAttribute ("action", buttonIdx + actionOffset)
--			ABT:DebugPrint ("Setting action on bar="..barIdx.." button="..buttonIdx.." to "..buttonIdx+actionOffset)
			height = button:GetHeight()
			width = button:GetWidth()
			button.status:ClearAllPoints()
			button.countdown:ClearAllPoints()
			button.tex:ClearAllPoints()
			button.spellName:ClearAllPoints()
			button.cooldown.noCooldownCount = nil -- for OmniCC
			button.timer:SetText(nil)
			local x, y, statusX, statusY
			if orientation == ABT_VERTICAL then
				x = outer_margin
				y = -1* (buttonIdx-1) * (height+margin) - outer_margin
				button.tex.orientation = "HORIZONTAL"
				button.tex:SetTexCoord(0, 0, 0, 1, 1, 0, 1, 1) -- standard texture orientation
				button.status:SetHeight(34)
				button.tex:SetHeight(34)
				button.status:SetWidth(length)
				if location == ABT_TOPRIGHT then
					button.timer.useMe = false
					button.tex.reverse = false
					button.status:SetPoint("TOPLEFT", button:GetName(), "TOPRIGHT")
					button.tex:SetPoint("LEFT", button.status:GetName(), "LEFT", 3, 0)
					button.countdown:SetPoint("LEFT", button.status:GetName(), "LEFT", 14, 0)
					button.spellName:SetPoint("LEFT", button.status:GetName(), "LEFT", 60, 0)
				elseif location == ABT_BOTTOMLEFT then
					button.timer.useMe = false
					button.tex.reverse = true
					button.status:SetPoint("TOPRIGHT", button:GetName(), "TOPLEFT")
					button.tex:SetPoint("RIGHT", button.status:GetName(), "RIGHT", -4, 0)
					button.countdown:SetPoint("RIGHT", button.status:GetName(), "RIGHT", -14, 0)
					button.spellName:SetPoint("RIGHT", button.status:GetName(), "RIGHT", -60, 0)
				elseif location == ABT_ONBUTTONS then
					button.cooldown.noCooldownCount = true -- for OmniCC
					button.timer.useMe = true
					button.status:Hide()
				end
			else
				x = (buttonIdx-1) * (width+margin) + outer_margin
				y = -outer_margin
				button.tex.orientation = "VERTICAL"
				button.tex:SetTexCoord(0, 1, 1, 1, 0, 0, 1, 0) -- rotate texture 90 degrees for vertical bar
				button.status:SetHeight(length)
				button.status:SetWidth(34)
				button.tex:SetWidth(34)
				if location == ABT_TOPRIGHT then
					button.timer.useMe = false
					button.tex.reverse = false
					button.status:SetPoint("BOTTOMLEFT", button:GetName(), "TOPLEFT")
					button.tex:SetPoint("BOTTOM", button.status:GetName(), "BOTTOM", 0, 3)
					button.countdown:SetPoint("BOTTOMLEFT", button.status:GetName(), "BOTTOMLEFT", 0, 14)
					button.spellName:SetPoint("BOTTOMLEFT", button.status:GetName(), "BOTTOMLEFT", 0, 32)
				elseif location == ABT_BOTTOMLEFT then
					button.timer.useMe = false
					button.tex.reverse = true
					button.status:SetPoint("TOPLEFT", button:GetName(), "BOTTOMLEFT")
					button.tex:SetPoint("TOP", button.status:GetName(), "TOP", 0, -4)
					button.countdown:SetPoint("TOPLEFT", button.status:GetName(), "TOPLEFT", 0, -14)
					button.spellName:SetPoint("TOPLEFT", button.status:GetName(), "TOPLEFT", 0, -32)
				elseif location == ABT_ONBUTTONS then
					button.cooldown.noCooldownCount = true -- for OmniCC
					button.timer.useMe = true
					button.status:Hide()
				end
			end
			button:ClearAllPoints()
			button:SetPoint ("TOPLEFT", x, y)
	
			if locked then
				bar.texture:Hide()
			else
				bar.texture:Show()
			end
		else
			button:Hide()
		end
	end
	
	-- layout
	if width ~= 0 and height ~= 0 then
		local barWidth
		local barHeight
		if orientation == ABT_VERTICAL then -- vertical
			barHeight = (height * buttonCount) + ((buttonCount - 1) * margin) + (outer_margin * 2)
			barWidth = width + (outer_margin * 2)
		else -- horizontal
			barHeight = height + (outer_margin * 2)
			barWidth = (width * buttonCount) + ((buttonCount - 1) * margin) + (outer_margin * 2)
		end
--		ABT:DebugPrint ("barHeight="..barHeight.." barWidth="..barWidth)
		bar:SetHeight(barHeight)
		bar:SetWidth(barWidth)
	end

	if inCombat and not InCombatLockdown() then
		bar:Hide()
	end	
	ABT_Bars[barIdx].needsRefresh = nil
	ABT:DebugTrace ("SetBar")
end

function ABT:UpdatePlayerTargets()
ABT:DebugTrace ("UpdatePlayerTargets", true)
	ABT:InitButtonLoop()
	repeat 
		barIdx, buttonIdx = ABT:GetNextButton()
		ABT:DebugPrint ("UPT==bar"..NS(barIdx).."button"..NS(buttonIdx))
		if barIdx and buttonIdx then
			if (ABT:GetValue(barIdx, buttonIdx, "target") == ABT_PARTY) then
				local playerTarget = ABT:GetValue (barIdx, buttonIdx, "playerTarget")
				local unitId = ABT:FindPlayerUnitId (playerTarget)
				local button = ABT_Buttons[barIdx][buttonIdx]
				button:SetAttribute ("unit", unitId)
			end
		end
	until buttonIdx == nil
	ABT:GetDebuffs()
	ABT:DebugTrace ("UpdatePlayerTargets")
end

function ABT:FindPlayerUnitId (playerTarget)
	local unitId = nil
	ABT:DebugPrint ("FindPlayerUnitId looking for: "..NS(playerTarget))
	if playerTarget ~= nil then
		local raidNum = GetNumRaidMembers()
		if raidNum > 0 then
			for i = 1, raidNum do
				if UnitName("raid"..i) == playerTarget then
					unitId = "raid"..i
					break
				end
			end
		else
			local partyNum = GetNumPartyMembers()
			for i = 1, partyNum do
				if UnitName("party"..i) == playerTarget then
					unitId = "party"..i
					break
				end
			end
		end
	end
	ABT:DebugPrint ("FindPlayerUnitId found "..NS(playerTarget).." as "..NS(unitId))
	return unitId
end

function ABT:GetPartyList ()
	local _table = {}
	local playerName = nil
	local raidNum = GetNumRaidMembers()
	if raidNum > 0 then
		for i = 1, raidNum do
			playerName = UnitName("raid"..i)
			if playerName then _table[playerName] = playerName end
		end
	else
		for i = 1, 4 do
			playerName = UnitName("party"..i)
			if playerName then _table[playerName] = playerName end
		end
	end
	return _table
end

function ABT:SetStatusBarRemaining (self, remaining)
	local value = remaining / self.tex.max
	if value > 1 then value = 1 end
	local length = ABT:GetValue(self.barIdx, nil, "length")
	if self.tex.orientation == "VERTICAL" then
		self.tex:SetHeight(value*length)
	else
		self.tex:SetWidth(value*length)
	end
end

function ABT:GetCastTime (button)
	local actionSlot = button:GetAttribute("action")	
	local actiontype, id, subType
	if actionSlot then
		actiontype, id, subType = GetActionInfo(actionSlot)
	end
	if actiontype == "spell" and id > 0 then
		local spellName = GetSpellName (id, BOOKTYPE_SPELL)
		local name, rank, icon, cost, isFunnel, powerType, castTime, minRange, maxRange = GetSpellInfo(spellName)
		ABT:DebugPrint (NS(actionSlot).." "..NS(actiontype).." "..NS(id).." "..NS(name).." Cast time= "..NS(castTime))
		return castTime
	end
end